//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//############################################################################
//# HISTORY  gps_nmea.c
//#
//# 27.06.2014 OG
//# - chg: NMEA_GetNewData() - zusaetzliche GPGGA Daten als Kommentar hinzugefuegt
//# - del: #include "../gps/mymath.h"
//#
//# 26.06.2014 OG - NEU
//# Code aus tracking.c / bluetooth.c / uart1.c zusammengefuehrt und strukturiert
//############################################################################


#include "../cpu.h"
#include <string.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdbool.h>
#include <stdlib.h>
#include <math.h>
#include "../main.h"

#include "../mk-data-structs.h"
#include "../lcd/lcd.h"
#include "../bluetooth/bluetooth.h"
#include "../gps/gps_nmea.h"


//----------------------------------------------------------------------------


NMEA_GPGGA_t    NMEA;

char NMEA_data_decode[RXD_BUFFER_SIZE];

uint8_t receiveNMEA = false;


#define MAX_POWER   10
#define getPower(x) (int32_t)pgm_read_dword(&powers[x])
static const int32_t    powers[MAX_POWER] PROGMEM = {1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000, 1000000000};



//############################################################################
//# Tools
//############################################################################

//--------------------------------------------------------------
//--------------------------------------------------------------
uint8_t NMEA_hexDigitToInt( uint8_t digit )
{
    if (digit >= '0' && digit <= '9') return digit - '0';
    if (digit >= 'a' && digit <= 'f') return digit - 'a' + 10;
    if (digit >= 'A' && digit <= 'F') return digit - 'A' + 10;
    return 0;
}



//--------------------------------------------------------------
// Trying to avoid floating point maths here. Converts a floating point string to an integer with a smaller unit
// i.e. floatStrToInt("4.5", 2) = 4.5 * 1E2 = 450
//--------------------------------------------------------------
int32_t NMEA_floatStrToInt( const char *s, int32_t power1 )
{
    char    *endPtr;
    int32_t  v;

    v = strtol(s, &endPtr, 10);

    if( *endPtr == '.' )
    {
        for (s = endPtr + 1; *s && power1; s++)
        {
            v = v * 10 + (*s - '0');
            --power1;
        }
    }

    if( power1 )
    {
        // Table to avoid multiple multiplications
        v = v * getPower(power1);
    }

    return v;
}



//############################################################################
//# Daten empfangen
//############################################################################

//--------------------------------------------------------------
//--------------------------------------------------------------
volatile uint8_t NMEA_receiveBT( void )
{
    static uint8_t line_flag = 1;
    static char*   ptr_write = bt_rx_buffer;
    uint32_t       timeout  = 400000;
    char received;


    while( !timeout == 0 )
    {
        if( bt_rx_ready == 1 )
            return true;

        uart_receive();
        if( fifo_is_empty( &in_fifo) ) timeout--;
        if( fifo_is_empty( &in_fifo) ) continue;

        timeout = 400000;
        fifo_read( &in_fifo, &received );

        if( bt_rx_ready == 0 )                              // Find starting point of packet
        {

            if( (received == '$') && (line_flag==1) )       // start == '$'
            {
                line_flag = 0;                              // New line has begun
                ptr_write = bt_rx_buffer;                   // Begin at start of buffer
                bt_rx_len = 0;
            }

            if( line_flag == 0 )
            {                                               // Are we receiving a line?
                *ptr_write = received;                      // Add current byte
                bt_rx_len++;

                // GPS Datensatzende
                if( received == '\n' )                      // End of MK-GPS or NMEA-line? end == '\n'
                {
                    line_flag = 1;                          // Yes, start new line
                    bt_rx_ready = 1;                        // Lock buffer until line has been processed

                   return true;
                }
            }

            ptr_write++;

            if( bt_rx_len == RXD_BUFFER_SIZE )              // Line too long? Try again
                line_flag = 1;

          } // end: if( bt_rx_ready == 0 )

    } // end: while( !timeout == 0 )

    return false;
}



//--------------------------------------------------------------
//--------------------------------------------------------------
uint8_t NMEA_decode( void )
{
    uint8_t  crc;
    uint8_t  tmpCRC = 0;
    uint8_t  lOk    = false;
    uint8_t  i;

    if( (bt_rx_ready == 1) && (bt_rx_len > 0) )
    {
         // Calculate checksum
         for( i = 1; i < bt_rx_len && bt_rx_buffer[i] != '*'; i++ )
         {
            tmpCRC ^= bt_rx_buffer[i];
         }

         if( bt_rx_len >= i + 3 )
         {
            crc = NMEA_hexDigitToInt( bt_rx_buffer[i + 1]) << 4 | NMEA_hexDigitToInt(bt_rx_buffer[i + 2] );

            if( crc == tmpCRC )
            {
                bt_rx_buffer[i] = 0;
                strcpy( NMEA_data_decode, &bt_rx_buffer[1] );           // Data without $, crc
                lOk = true;
            }
        }
    } // end: if( (bt_rx_ready == 1) && (bt_rx_len > 0) )


    //if (rx_timeout < RX_TIME_OLD) wi232RX = 1;
    bt_rx_ready = 0;                                                // Unlock buffer, next NMEA string can be received

    return lOk;
}



//############################################################################
//# Daten parsen
//############################################################################

//--------------------------------------------------------------
// NMEA latitudes are in the form ddmm.mmmmm, we want an integer in 1E-7 degree steps
//--------------------------------------------------------------
int32_t NMEA_getLatitude( const char *s, const char *NS)
{
    int32_t deg;
    int32_t min;

    //lcdx_puts_at(0,5,NS,0,0,0);
    deg = (s[0] - '0') * 10 + s[1] - '0';                           // First 2 chars are full degrees
    min = NMEA_floatStrToInt( &s[2], 6) / 6;                         // Minutes * 1E5 * 100 / 60 = Minutes * 1E6 / 6 = 1E-7 degree steps

    deg = deg * 10000000 + min;

    if( *NS == 'S' )
    {
        deg = -deg;
    }

    return deg;
}



//--------------------------------------------------------------
// NMEA longitudes are in the form dddmm.mmmmm, we want an integer in 1E-7 degree steps
//--------------------------------------------------------------
int32_t NMEA_getLongitude( const char *s, const char *WE)
{
    int32_t deg;
    int32_t min;

    //lcdx_puts_at(10,5,WE,0,0,0);
    deg = ((s[0] - '0') * 10 + s[1] - '0') * 10 + s[2] - '0';       // First 3 chars are full degrees
    min = NMEA_floatStrToInt( &s[3], 6) / 6;                         // Minutes * 1E5 * 100 / 60 = Minutes * 1E6 / 6 = 1E-7 degree steps

    deg = deg * 10000000 + min;

    if( *WE == 'W' )
    {
        deg = -deg;
    }

    return deg;
}



//--------------------------------------------------------------
//--------------------------------------------------------------
void NMEA_getTime( const char *s)
{
    uint8_t sem = 0;
    uint8_t i;

    for( i=0; i<6; i++ )
    {
        NMEA.Time[sem++] = s[i];

        if( i==1 || i==3)
            NMEA.Time[sem++] = ':';

    }
    NMEA.Time[sem] = '\0';
}




//$GPGGA,191410.000,4735.5634,N,00739.3538,E,1,04,4.4,351.5,M,48.0,M,,*45
//           ^      ^           ^            ^ ^  ^   ^       ^
//           |      |           |            | |  |   |       |
//           |      |           |            | |  |   |       Höhe Geoid minus
//           |      |           |            | |  |   |       Höhe Ellipsoid (WGS84)
//           |      |           |            | |  |   |       in Metern (48.0,M)
//           |      |           |            | |  |   |
//           |      |           |            | |  |   Höhe über Meer (über Geoid)in Metern (351.5,M)
//           |      |           |            | |  |
//           |      |           |            | |  HDOP (horizontal dilution
//           |      |           |            | |  of precision) Genauigkeit
//           |      |           |            | |
//           |      |           |            | Anzahl der erfassten Satelliten
//           |      |           |            |
//           |      |           |            Qualität der Messung
//           |      |           |            (0 = ungültig)
//           |      |           |            (1 = GPS)
//           |      |           |            (2 = DGPS)
//           |      |           |            (6 = geschätzt nur NMEA-0183 2.3)
//           |      |           |
//           |      |           Längengrad
//           |      |
//           |      Breitengrad
//           |
//           Uhrzeit

//--------------------------------------------------------------
// NMEA_GetNewData()
//
// ehemals: Tracking_NMEA() in tracking.c
//--------------------------------------------------------------
void NMEA_GetNewData( void )
{
    char *token;
    char *param1;
    char *param2;


    if( NMEA_decode() )
    {
        token = strtok( (char*)NMEA_data_decode, ",");

        if( !strcmp(token, "GPGGA") )
        {
            NMEA.Counter++;                                                 // Zaehler erfolgreiche GPGGA Pakete

            //-------------------------------------------------------------------------
            // $GPGGA,220613.400,4843.5080,N,00922.9583,E,1,7,2.23,287.1,M,48.0,M,,
            //
            // http://www.trimble.com/OEM_ReceiverHelp/V4.44/en/NMEA-0183messages_GGA.html
            //
            // Chipset MTK:
            // http://guides.machinescience.org/mod/book/print.php?id=1368
            //-------------------------------------------------------------------------
            NMEA_getTime( strtok(0, ".") );                                 // Zeit

            strtok(0, ",");                                                 // Skip Rest von der Zeit

            param1 = strtok(0, ",");                                        // Latitude
            param2 = strtok(0, ",");
            NMEA.Latitude = NMEA_getLatitude( param1,param2 );              // -> Latitude (N)

            param1 = strtok(0, ",");                                        // Longitude
            param2 = strtok(0, ",");
            NMEA.Longitude = NMEA_getLongitude( param1,param2 );            // -> Longitude (E)

            NMEA.SatFix = atoi( strtok(0, ","));                            // Qualitaet  (Signal valid? - Position Fix Indicator)

            NMEA.SatsInUse = atoi( strtok(0, ","));                         // Anzahl Sats

            NMEA.HDOP = NMEA_floatStrToInt( strtok(0, ","),1);              // HDOP

            NMEA.Altitude = NMEA_floatStrToInt( strtok(0, ","), 1);         // Altitude in Meter

            /*
            //-----------------------
            // weitere GPGGA Daten
            //-----------------------
            param1 = strtok(0, ",");
            //lcdx_printf_at_P( 0, 1, MNORMAL, 0,1, PSTR("%s"), param1 );   // Altitude Unit "M"

            param1 = strtok(0, ",");
            //lcdx_printf_at_P( 2, 1, MNORMAL, 0,1, PSTR("%s"), param1 );   // Geoid separation eg. "47.6"

            param1 = strtok(0, ",");
            //lcdx_printf_at_P( 8, 1, MNORMAL, 0,1, PSTR("%s"), param1 );   // Geoid separation Unit "M"; geoid separation measured in meters

            param1 = strtok(0, ",");
            lcdx_printf_at_P( 0, 1, MNORMAL, 0,1, PSTR("%s"), param1 );     // Age of differential GPS data record in seconds, Type 1 or Type 9. Null field when DGPS is not used.

            param1 = strtok(0, ",");
            lcdx_printf_at_P( 8, 1, MNORMAL, 0,1, PSTR("%s"), param1 );     // DGPS Reference station ID, range 0000-4095. A null field when any reference station ID is selected and no corrections are received.
            */
        }
    }
}

